// ** THIS FILE IS GENERATED **
// ** use generateConvert.php to update this file **


#pragma once

constexpr const char* ACTION_NAMES[] = {  // Action to string conversion map
        "new-file",
        "open",
        "annotate-pdf",
        "save",
        "save-as",
        "export-as-pdf",
        "export-as",
        "print",
        "quit",
        "arrange-selection-order",
        "undo",
        "redo",
        "cut",
        "copy",
        "paste",
        "search",
        "select-all",
        "delete",
        "move-selection-layer-up",
        "move-selection-layer-down",
        "rotation-snapping",
        "grid-snapping",
        "preferences",
        "paired-pages-mode",
        "paired-pages-offset",
        "presentation-mode",
        "fullscreen",
        "show-sidebar",
        "show-toolbar",
        "set-layout-vertical",
        "set-layout-right-to-left",
        "set-layout-bottom-to-top",
        "set-columns-or-rows",
        "manage-toolbar",
        "customize-toolbar",
        "show-menubar",
        "zoom-in",
        "zoom-out",
        "zoom-100",
        "zoom-fit",
        "zoom",
        "goto-first",
        "goto-previous",
        "goto-page",
        "goto-next",
        "goto-last",
        "goto-next-annotated-page",
        "goto-previous-annotated-page",
        "new-page-before",
        "new-page-after",
        "new-page-at-end",
        "duplicate-page",
        "move-page-towards-beginning",
        "move-page-towards-end",
        "append-new-pdf-pages",
        "configure-page-template",
        "delete-page",
        "paper-format",
        "paper-background-color",
        "select-tool",
        "select-default-tool",
        "tool-draw-shape-recognizer",
        "tool-draw-rectangle",
        "tool-draw-ellipse",
        "tool-draw-arrow",
        "tool-draw-double-arrow",
        "tool-draw-coordinate-system",
        "tool-draw-line",
        "tool-draw-spline",
        "setsquare",
        "compass",
        "tool-pen-size",
        "tool-pen-line-style",
        "tool-pen-fill",
        "tool-pen-fill-opacity",
        "tool-eraser-size",
        "tool-eraser-type",
        "tool-highlighter-size",
        "tool-highlighter-fill",
        "tool-highlighter-fill-opacity",
        "tool-select-pdf-text-marker-opacity",
        "toggle-touch-drawing",
        "audio-record",
        "audio-pause-playback",
        "audio-stop-playback",
        "audio-seek-forwards",
        "audio-seek-backwards",
        "select-font",
        "font",
        "tex",
        "plugin-manager",
        "help",
        "demo",
        "about",
        "tool-size",
        "tool-fill",
        "tool-fill-opacity",
        "tool-color",
        "select-color",
        "layer-show-all",
        "layer-hide-all",
        "layer-new-above-current",
        "layer-new-below-current",
        "layer-copy",
        "layer-move-up",
        "layer-move-down",
        "layer-delete",
        "layer-merge-down",
        "layer-rename",
        "layer-goto-next",
        "layer-goto-previous",
        "layer-goto-top",
        "layer-active",
        "position-highlighting"};
